# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import Plugin
from canto_next.encoding import locale_enc
from canto_next.hooks import on_hook

from .command import CommandHandler, register_commands, register_arg_types
from .taglist import TagList
from .input import InputBox
from .text import InfoBox
from .widecurse import wsize, set_redisplay_callback, set_getc, raw_readline
from .locks import sync_lock

from threading import Lock
import traceback
import readline
import logging
import curses
import time
import os

log = logging.getLogger("SCREEN")

# The Screen class handles the layout of multiple sub-windows on the main
# curses window. It's also the top-level gui object, so it handles calls to
# refresh the screen, get input, and curses related console commands, like
# "color".

# There are two types of windows that the Screen class handles. The first are
# normal windows (in self.tiles). These windows are all tiled in a single
# layout (determined by self.layout and self.fill_layout()) and rendered first.

# The other types are floats that are rendered on top of the window layout.
# These floats are all independent of each other.

# The Screen class is also in charge of honoring the window specific
# configuration options. Like window.{maxwidth,maxheight,float}.

class ScreenPlugin(Plugin):
    pass

class Screen(CommandHandler):
    def __init__(self, callbacks, types = [InputBox, TagList]):
        CommandHandler.__init__(self)

        self.plugin_class = ScreenPlugin
        self.update_plugin_lookups()

        self.callbacks = callbacks
        self.layout = "default"

        self.window_types = types

        self.input_box = None
        self.pseudo_input_box = None
        self.in_readline = False
        self.in_readline_resize = False

        self.stdscr = curses.initscr()
        if self.curses_setup() < 0:
            return -1

        self.pseudo_input_box = curses.newpad(1,1)

        self.pseudo_input_box.keypad(1)
        self.pseudo_input_box.nodelay(1)
        self.input_lock = Lock()

        set_redisplay_callback(self.readline_redisplay)
        set_getc(self.readline_getc)

        # See Python bug 2675, readline + curses
        os.unsetenv('LINES')
        os.unsetenv('COLUMNS')

        self.floats = []
        self.tiles = []
        self.windows = []

        self.subwindows()

        args = {
            "color_name" : ("[color name] Either a pair number (0-255, >8 ignored on 8 color terminals), a default fore/background (deffg, defbg), or an arbitrary name to be used in themes (unread, pending, etc.)", self.type_color_name),
            "fg-color" : ("[fg-color] Foreground color", self.type_color),
            "bg-color" : ("[bg-color] Background color (optional)\n\nNamed colors: white black red yellow green blue magenta pink\nNumeric colors: 1-256", self.type_color),
            "style" : ("[style] Curses style (normal, bold, dim, reverse, standout, underline)", self.type_style),
        }

        cmds = {
            "color": (self.cmd_color, ["color_name", "fg-color", "bg-color"],
"""Change the color palette.

Most like you want to use this to change a color used in the theme. For example,

   :color unread green

Will change the color of unread items to green, with the default background. The list of names used in the default theme are:

    unread
    read
    selected
    marked
    pending
    error
    reader_quote
    reader_link
    reader_image_link
    reader_italics
    enum_hints

You can also change the defaults

    :color deffg blue
    :color defbg white

Which will be used anywhere a color pair doesn't have an explicit foreground/background.

Lastly you can change the color pairs themselves. This isn't recommended, they're initialized so each available color is available with the default background. If you change these pairs, the named colors above may not make any sense (i.e. green really turns on the color pair set aside for green, so if you change that pair to actually be yellow, don't expect this command to figure it out).

    :color 1 white red

Arguments:"""

),
        "style": (self.cmd_style, ["color_name", "style"],
"""Change the curses style of a named color.

For example,

    :style selected underline

The names used in the default theme are:

    unread
    read
    selected
    marked
    pending
    error
    reader_quote
    reader_link
    reader_image_link
    reader_italics
    enum_hints

Changing other colors (numeric pairs or deffg/defbg) will have no effect as these are separate from the built in curses color system.

Arguments:"""),
        }

        register_arg_types(self, args)
        register_commands(self, cmds, "Theme")

        on_hook("curses_opt_change", self.screen_opt_change)

    # Wrap curses.curs_set in exception handler
    # because we don't really care if it's displayed
    # on terminals that don't like it.

    def curs_set(self, n):
        try:
            curses.curs_set(n)
        except:
            pass

    # Do initial curses setup. This should only be done on init, or after
    # endwin() (i.e. resize).

    def curses_setup(self):
        if self.in_readline:
            self.pseudo_input_box.keypad(0)
            self.curs_set(1)
        elif self.pseudo_input_box:
            self.pseudo_input_box.nodelay(1)
            self.pseudo_input_box.keypad(1)
            self.curs_set(0)
        else:
            self.curs_set(0)

        try:
            curses.cbreak()
            curses.noecho()
            curses.start_color()
            curses.use_default_colors()
            curses.typeahead(-1)
            curses.halfdelay(5)
        except Exception as e:
            log.error("Curses setup failed: %s" % e.msg)
            return -1

        self.height, self.width = self.stdscr.getmaxyx()
        self.height = int(self.height)
        self.width = int(self.width)

        color_conf = self.callbacks["get_opt"]("color")

        for i in range(0, curses.COLOR_PAIRS):
            if str(i) not in color_conf:
                continue

            color = color_conf[str(i)]

            if type(color) == int:
                fg = color
                bg = color_conf['defbg']
            else:
                if 'fg' in color:
                    fg = color['fg']
                else:
                    fg = color_conf['deffg']

                if 'bg' in color:
                    bg = color['bg']
                else:
                    bg = color_conf['defbg']

            try:
                curses.init_pair(i, fg, bg)
                log.debug("color pair %s : %s %s", i, fg, bg)
            except:
                log.debug("color pair failed!: %d fg: %d bg: %d", 
                        i + 1, fg, bg)
        return 0

    def screen_opt_change(self, conf):
        # Require resize even to re-init curses and colors.
        if "color" in conf:
            self.callbacks["set_var"]("needs_resize", True)

        for key in list(conf.keys()):
            if type(conf[key]) == dict and "window" in conf[key]:
                self.callbacks["set_var"]("needs_resize", True)
                break

    # _subw_size functions enforce the height and width of windows.
    # It returns the minimum of:
    #       - The maximum size (given by layout)
    #       - The requested size (given by the class)
    #       - The configured size (given by the config)

    def _subw_size_height(self, ci, height):
        window_conf = self.callbacks["get_opt"](ci.get_opt_name() + ".window")

        if not window_conf["maxheight"]:
            window_conf["maxheight"] = height
        req_height = ci.get_height(height)

        return min(height, window_conf["maxheight"], req_height)

    def _subw_size_width(self, ci, width):
        window_conf = self.callbacks["get_opt"](ci.get_opt_name() + ".window")

        if not window_conf["maxwidth"]:
            window_conf["maxwidth"] = width
        req_width = ci.get_width(width)

        return min(width, window_conf["maxwidth"], req_width)

    # _subw_layout_size will return the total size of layout
    # in either height or width where layout is a list of curses
    # pads, or sublists of curses pads.

    def _subw_layout_size(self, layout, dim):

        # Grab index into pad.getmaxyx()
        if dim == "width":
            idx = 1
        elif dim == "height":
            idx = 0
        else:
            raise Exception("Unknown dim: %s" % dim)

        sizes = []
        for x in layout:
            if hasattr(x, "__iter__"):
                sizes.append(self._subw_layout_size(x, dim))
            else:
                sizes.append(x.pad.getmaxyx()[idx] - 1)

        return max(sizes)

    # Translate the layout into a set of curses pads given
    # a set of coordinates relating to how they're mapped to the screen.

    def _subw_init(self, ci, top, left, height, width):

        # Height - 1 because start + height = line after bottom.

        bottom = top + (height - 1)
        right = left + (width - 1)

        # lambda this up so that subwindows truly have no idea where on the
        # screen they are, only their dimensions, but can still selectively
        # refresh their portion of the screen.

        refcb = lambda : self.refresh_callback(ci, top, left, bottom, right)

        # Callback to allow windows to know if they're floating. This is
        # important because floating windows are only rendered up to their
        # last cursor position, despite being given a maximal window.

        floatcb = lambda : ci in self.floats

        # Use coordinates and dimensions to determine where borders
        # are needed. This is independent of whether there are actually
        # windows there.

        # NOTE: These should only be honored if the window is non-floating.
        # Floating windows are, by design, given a window the size of the
        # entire screen, but only actually written lines are drawn.

        window_conf = self.callbacks["get_opt"](ci.get_opt_name() + ".window")

        if window_conf['border'] == "smart":
            top_border = top != 0
            bottom_border = bottom != (self.height - 1)
            left_border = left != 0
            right_border = right != (self.width - 1)

            if ci in self.floats:
                if "top" in window_conf['align']:
                    bottom_border = True
                if "bottom" in window_conf['align']:
                    top_border = True

        elif window_conf['border'] == "full":
            top_border, bottom_border, left_border, right_border = (True,) * 4

        elif window_conf['border'] == "none":
            top_border, bottom_border, left_border, right_border = (False,) * 4

        bordercb = lambda : (top_border, left_border, bottom_border, right_border)

        # Height + 1 to account for the last curses pad line
        # not being fully writable.

        log.debug("h: %s w: %s", self.height, self.width)
        log.debug("h: %s w: %s", height, width)
        pad = curses.newpad(height + 1, width)

        # Pass on callbacks we were given from CantoCursesGui
        # plus our own.

        callbacks = self.callbacks.copy()
        callbacks["refresh"] = refcb
        callbacks["border"] = bordercb
        callbacks["floating"] = floatcb
        callbacks["input"] = self.input_callback
        callbacks["die"] = self.die_callback
        callbacks["pause_interface" ] = self.pause_interface_callback
        callbacks["unpause_interface"] = self.unpause_interface_callback
        callbacks["add_window"] = self.add_window_callback

        ci.init(pad, callbacks)

    # Layout some windows into the given space, stacking with
    # orientation horizontally or vertically.

    def _subw(self, layout, top, left, height, width, orientation):
        immediates = []
        cmplx = []
        sizes = [0] * len(layout)

        # Separate windows in to two categories:
        # immediates that are defined as base classes and
        # cmplx which are lists for further processing (iterables)

        for i, unit in enumerate(layout):
            if hasattr(unit, "__iter__"):
                cmplx.append((i, unit))
            else:
                immediates.append((i,unit))

        # Units are the number of windows we'll have
        # to split the area with.

        units = len(layout)

        # Used, the amounts of space already used.
        used = 0

        for i, unit in immediates:
            # Get the size of the window from the class.
            # Each class is given, as a maximum, the largest
            # possible slice we can *guarantee*.

            if orientation == "horizontal":
                size = self._subw_size_width(unit, int((width - used) / units))
            else:
                size = self._subw_size_height(unit, int((height - used) / units))

            used += size

            sizes[i] = size

            # Subtract so that the next run only divides
            # the remaining space by the number of units
            # that don't have space allocated.

            units -= 1

        # All of the immediates have been allocated for.
        # So now only the cmplxs are vying for space.

        units = len(cmplx)

        for i, unit in cmplx:
            offset = sum(sizes[0:i])

            # Recursives call this function, alternating
            # the orientation, for the space we can guarantee
            # this set of windows.

            if orientation == "horizontal":
                available = int((width - used) / units)
                r = self._subw(unit, top, left + offset,\
                        height, available, "vertical")
                sizes[i] = self._subw_layout_size(r, "width")
            else:
                available = int((height - used) / units)
                r = self._subw(unit, top + offset, left,\
                        available, width, "horizontal")
                sizes[i] = self._subw_layout_size(r, "height")

            used += sizes[i]
            units -= 1

        # Now that we know the actual sizes (and thus locations) of
        # the windows, we actually setup the immediates.

        for i, ci in immediates:
            offset = sum(sizes[0:i])
            if orientation == "horizontal":
                self._subw_init(ci, top, left + offset,
                        height, sizes[i])
            else:
                self._subw_init(ci, top + offset, left,
                        sizes[i], width)
        return layout

    # The fill_layout() function takes a list of active windows and generates a
    # list based layout. The depth of a window in the list determines its
    # orientation.
    #
    #   Example return: [ Window1, Window2 ]
    #       - Window1 on top of Window2, each taking half of the vertical space.
    #
    #   Example return: [ [ Window1, Window2 ], Window 3 ]
    #       - Window1 left of Window2 each taking half of the horizontal space,
    #           and whatever vertical space left by Window3, because Window3 is
    #           shallower than 1 or 2, so it's size is evaluated first and the
    #           remaining given to the [ Window1, Window2 ] horizontal layout.
    #
    #   Example return: [ [ [ [ Window1 ] ], Window2 ], Window3 ]
    #       - Same as above, except because Window1 is deeper than Window2 now,
    #           Window2's size is evaluated first and Window1 is given all of 
    #           the remaining space.
    #
    #   NOTE: Floating windows are not handled in the layout, this is solely for
    #   the tiling bottom layer of windows.

    def fill_layout(self, layout, windows):
        inputs = [ w for w in windows if w.is_input() ]
        if inputs:
            self.input_box = inputs[0]
        else:
            self.input_box = None

        # Simple stacking, even distribution between all windows.
        if layout == "hstack":
            return windows
        elif layout == "vstack":
            return [ windows ]
        else:
            aligns = { "top" : [], "bottom" : [], "left" : [], "right" : [],
                            "neutral" : [] }

            # Separate windows by alignment.
            for w in windows:
                align = self.callbacks["get_opt"]\
                        (w.get_opt_name() + ".window.align")

                # Move taglist deeper so that it absorbs any
                # extra space left in the rest of the layout.

                if w.get_opt_name() == "taglist":
                    aligns[align].append([[w]])
                else:
                    aligns[align].append(w)

            horizontal = aligns["left"] + aligns["neutral"] + aligns["right"]
            return aligns["top"] + [horizontal] + aligns["bottom"]

    # subwindows() is the top level window generator. It handles both the bottom
    # level tiled window layout as well as the floats.

    def subwindows(self):

        # Cleanup any window objects that will be destroyed.
        for w in self.windows:
            log.debug("die to %s", w)
            w.die()

        self.floats = []
        self.tiles = []
        self.windows = []

        # Instantiate new windows, separating them into
        # floating and tiling windows.

        for wt in self.window_types:
            w = wt()
            optname = w.get_opt_name()
            flt = self.callbacks["get_opt"](optname + ".window.float")
            if flt:
                self.floats.append(w)
            else:
                self.tiles.append(w)
            self.windows.append(w)

        # Focused window will no longer exist.
        self.focused = None

        # Init tiled windows.
        l = self.fill_layout(self.layout, self.tiles)
        self._subw(l, 0, 0, self.height, self.width, "vertical")

        # Init floating windows.
        for f in self.floats: 
            align = self.callbacks["get_opt"]\
                    (f.get_opt_name() + ".window.align")
            height = self._subw_size_height(f, self.height)
            width = self._subw_size_width(f, self.width)

            top = 0
            if align.startswith("bottom"):
                top = self.height - height

            left = 0
            if align.endswith("right"):
                left = self.width - width

            self._subw_init(f, top, left, height, width)

        # Default to giving first window focus.
        self._focus_abs(0)

    def refresh_callback(self, c, t, l, b, r):
        if c in self.floats:
            b = min(b, t + c.pad.getyx()[0])
        c.pad.noutrefresh(0, 0, t, l, b, r)

    def input_callback(self, prompt, completions=True):
        # Setup subedit
        self.curs_set(1)

        self.callbacks["set_var"]("input_do_completions", completions)
        self.callbacks["set_var"]("input_prompt", prompt)

        self.input_box.reset()
        self.input_box.refresh()
        curses.doupdate()

        self.pseudo_input_box.keypad(0)

        self.in_readline = True
        r = raw_readline()
        self.in_readline = False

        if self.in_readline_resize:
            self.in_readline_resize = False
            self.resize()

        if not r:
            r = ""

        self.pseudo_input_box.keypad(1)

        # Only add history for commands, not other prompts
        if completions:
            readline.add_history(r)

        self.callbacks["set_var"]("input_prompt", "")
        self.input_box.reset()
        self.input_box.refresh()
        curses.doupdate()

        self.curs_set(0)
        return r

    def die_callback(self, window):
        sync_lock.acquire_write()

        # Remove window from both window_types and the general window list

        try:
            idx = self.windows.index(window)
        except:
            pass
        else:
            # Call the window's die function
            window.die()

            del self.windows[idx]
            del self.window_types[idx]

            # Regenerate layout with remaining windows.
            self.subwindows()

            self.refresh()

            # Force a doupdate because refresh doesn't, but we have possibly
            # uncovered part of the screen that isn't handled by any other window.

            curses.doupdate()

        sync_lock.release_write()

    # The pause interface callback keeps the interface from updating. This is
    # useful if we have to temporarily surrender the screen (i.e. text browser).

    # NOTE: This does not affect signals so even while "paused", c-c continues
    # to take things like SIGWINCH which will be interpreted on wakeup.

    def pause_interface_callback(self):
        log.debug("Pausing interface.")
        sync_lock.acquire_write()

        curses.raw()

        self.input_lock.acquire()

        curses.endwin()

    def unpause_interface_callback(self):
        log.debug("Unpausing interface.")
        self.input_lock.release()

        # All of our window information could be stale.
        self.resize()
        sync_lock.release_write()

    def add_window_callback(self, cls):
        sync_lock.acquire_write()

        self.window_types.append(cls)

        self.subwindows()

        # Focus new window
        self._focus_abs(0)

        self.refresh()
        self.redraw()

        sync_lock.release_write()

    def _readline_redisplay(self):
        self.input_box.refresh()
        curses.doupdate()

    def _readline_getc(self):
        do_comp = self.callbacks["get_var"]("input_do_completions")

        # Don't flush, because readline loses keys.
        r = self.get_key(False)

        if r == curses.KEY_BACKSPACE:
            r = ord("\b")
        elif r == curses.KEY_RESIZE:
            self.in_readline_resize = True
            return
        elif chr(r) == '\t' and do_comp:
            self.input_box.rotate_completions()
            return

        comp = self.input_box.break_completion()
        if comp:
            log.debug("inserting: %s", comp)
            readline.insert_text(comp)

        log.debug("KEY: %s", r)
        return r

    def _exception_wrap(self, fn, *args):
        r = None
        try:
            r = fn(*args)
        except:
            log.error("".join(traceback.format_exc()))
        return r

    def readline_redisplay(self, *args):
        return self._exception_wrap(self._readline_redisplay, *args)
    def readline_complete(self, *args):
        return self._exception_wrap(self._readline_complete, *args)
    def readline_display_matches(self, *args):
        return self._exception_wrap(self._readline_display_matches, *args)
    def readline_getc(self, *args):
        return self._exception_wrap(self._readline_getc, *args)

    # Refresh operates in order, which doesn't matter for top level tiled
    # windows, but this ensures that floats are ordered such that the last
    # floating window is rendered on top of all others.

    def refresh(self):
        for c in self.tiles + self.floats:
            c.refresh()

    def redraw(self):
        for c in self.tiles + self.floats:
            c.redraw()
        curses.doupdate()

    # Typical curses resize, endwin and re-setup.
    def resize(self):
        try:
            curses.endwin()
        except:
            pass

        self.stdscr.refresh()

        self.curses_setup()
        self.subwindows()

        self.refresh()
        self.redraw()

    def _focus_abs(self, idx):
        focus_order = self.tiles + self.floats
        focus_order.reverse()
        l = len(focus_order)

        if idx < 0:
            idx = -1 * (idx % l)
        else:
            idx %= l

        self._focus(focus_order[idx])

    def _focus(self, win):
        self.focused = win
        log.debug("Focusing window (%s)", self.focused)

    def type_color(self):
        colors = {
            'white' : curses.COLOR_WHITE,
            'black' : curses.COLOR_BLACK,
            'red' : curses.COLOR_RED,
            'blue' : curses.COLOR_BLUE,
            'green' : curses.COLOR_GREEN,
            'yellow' : curses.COLOR_YELLOW,
            'cyan' : curses.COLOR_CYAN,
            'magenta' : curses.COLOR_MAGENTA,
            'pink' : curses.COLOR_MAGENTA,
            "default" : -1,
        }
        def c(x):
            if x == '':
                return (True, -1)

            if x in colors:
                return (True, colors[x])
            try:
                x = int(x)
                if x in list(range(-1, 256)):
                    return (True, x)
                return (False, None)
            except:
                return (False, None)
        return (list(colors.keys()), c)

    def type_color_name(self):
        color_conf = self.callbacks["get_opt"]("color")

        completions = []
        for key in color_conf:
            try:
                pair = int(key)
                continue
            except:
                completions.append(key)

        return (completions, lambda x : (True, x))

    def cmd_color(self, idx, fg, bg):
        conf = self.callbacks["get_conf"]()

        log.debug(":COLOR IDX: %s", idx)
        if idx in ['deffg', 'defbg']:
            conf["color"][idx] = fg  # Ignore second color pair
        else:
            try:
                pair = int(idx)
                color = {}
                if fg != -1:
                    color['fg'] = fg
                if bg != -1:
                    color['bg'] = bg
                conf["color"][idx] = color
            except:
                conf["color"][idx] = fg + 1 # +1 since the color pairs are offset

        log.debug("color %s set: %s", idx, conf["color"][idx])

        self.callbacks["set_conf"](conf)

        # Cause curses to be re-init'd
        self.resize()

    def type_style(self):
        styles = [ "normal", "bold", "dim", "reverse", "standout", "underline" ]
        return (styles, lambda x: (x in styles, x))

    def cmd_style(self, name, style):
        conf = self.callbacks["get_opt"]("style")

        styles = {
            "bold" : "%B",
            "normal" : "",
            "dim" : "%D",
            "standout" : "%S",
            "reverse" : "%R",
            "underline" : "%U",
        }

        conf[name] = styles[style]

        log.debug("style %s set: %s", name, style)

        self.callbacks["set_opt"]("style", conf)

        self.resize()

    def get_focus_list(self):

        # self.focused might be None, if this is getting called during a
        # subwindows() call.

        if self.focused:
            return [ self, self.focused ]
        return [ self ]

    def get_key(self, flush=True):
        while True:
            self.input_lock.acquire()
            try:
                r = self.pseudo_input_box.get_wch()
            except Exception as e:
                r = self.pseudo_input_box.getch()
            self.input_lock.release()
            if r != -1:
                break

        if flush and r != curses.KEY_RESIZE:
            curses.flushinp()

        if type(r) == str:
            r = ord(r)
        return r

    def exit(self):
        curses.endwin()

    def get_opt_name(self):
        return "screen"



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.rwlock import RWLock

config_lock = RWLock('config_lock')

# This lock can be held with write to keep sync operations from happening.
sync_lock = RWLock("global sync lock")



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import Plugin
from canto_next.hooks import on_hook, remove_hook, unhook_all

from .command import register_commands, register_arg_types, unregister_all, _int_range
from .theme import prep_for_display
from .html import htmlparser
from .text import TextBox
from .tagcore import tag_updater
from .color import cc

import traceback
import logging
import re

log = logging.getLogger("READER")

class ReaderPlugin(Plugin):
    pass

class Reader(TextBox):
    def init(self, pad, callbacks):
        TextBox.init(self, pad, callbacks)

        self.quote_rgx = re.compile("[\\\"](.*?)[\\\"]")
        on_hook("curses_opt_change", self.on_opt_change, self)
        on_hook("curses_var_change", self.on_var_change, self)

        args = {
            "link-list" : ("", self.type_link_list),
        }

        cmds = {
            "goto" : (self.cmd_goto, ["link-list"], "Goto a specific link"),
            "destroy" : (self.cmd_destroy, [], "Destroy this window"),
            "show-links" : (self.cmd_show_links, [], "Toggle link list display"),
            "show-summary" : (self.cmd_show_desc, [], "Toggle summary display"),
            "show-enclosures" : (self.cmd_show_encs, [], "Toggle enclosure list display")
        }

        register_arg_types(self, args)
        register_commands(self, cmds, "Reader")

        self.plugin_class = ReaderPlugin
        self.update_plugin_lookups()

    def die(self):
        unhook_all(self)
        unregister_all(self)

    def on_opt_change(self, change):
        if "reader" not in change:
            return

        for opt in ["show_description", "enumerate_links", "show_enclosures"]:
            if opt in change["reader"]:
                self.callbacks["set_var"]("needs_refresh", True)
                self.callbacks["release_gui"]()
                return

    def on_attributes(self, attributes):
        sel = self.callbacks["get_var"]("reader_item")
        if sel and sel.id in attributes:
            remove_hook("curses_attributes", self.on_attributes)
            self.callbacks["set_var"]("needs_refresh", True)
            self.callbacks["release_gui"]()

    def on_var_change(self, variables):
        # If we've been instantiated and unfocused, and selection changes,
        # we need to be redrawn.

        if "selected" in variables and variables["selected"]:
            self.callbacks["set_var"]("reader_item", variables["selected"])
            self.callbacks["set_var"]("needs_refresh", True)
            self.callbacks["release_gui"]()

    def type_link_list(self):
        domains = { 'all' : self.links }
        syms = { 'all' : { '*' : range(0, len(self.links)) } }

        fallback = []
        if len(self.links):
            fallback = [ self.links[0] ]

        return (None, lambda x:_int_range("link", domains, syms, fallback, x))

    def update_text(self):
        reader_conf = self.callbacks["get_opt"]("reader")

        s = "No selected story.\n"
        extra_content = ""

        sel = self.callbacks["get_var"]("reader_item")
        if sel:
            self.links = [("link",sel.content["link"],"mainlink")]

            s = "%B" + prep_for_display(sel.content["title"]) + "%b\n"

            # Make sure the story has the most recent info before we check it.
            sel.sync()

            # We use the description for most reader content, so if it hasn't
            # been fetched yet then grab that from the server now and setup
            # a hook to get notified when sel's attributes are changed.

            l = ["description", "content", "links", "media_content",
                    "enclosures"]

            for attr in l:
                if attr not in sel.content:
                    tag_updater.request_attributes(sel.id, l)
                    s += "%BWaiting for content...%b\n"
                    on_hook("curses_attributes", self.on_attributes, self)
                    break
            else:
                # Grab text content over description, as it's likely got more
                # information.

                mainbody = sel.content["description"]
                if "content" in sel.content:
                    for c in sel.content["content"]:
                        if "type" in c and "text" in c["type"]:
                            mainbody = c["value"]

                # Add enclosures before HTML parsing so that we can add a link
                # and have the remaining link logic pick it up as normal.

                if reader_conf['show_enclosures']:
                    parsed_enclosures = []

                    if sel.content["links"]:
                        for lnk in sel.content["links"]:
                            if 'rel' in lnk and 'href' in lnk and lnk['rel'] == 'enclosure':
                                if 'type' not in lnk:
                                    lnk['type'] = 'unknown'
                                parsed_enclosures.append((lnk['href'], lnk['type']))

                    if sel.content["media_content"] and 'href' in sel.content["media_content"]:
                        if 'type' not in sel.content["media_content"]:
                            sel.content['media_content']['type'] = 'unknown'
                        parsed_enclosures.append((sel.content["media_content"]['href'],\
                                    sel.content["media_content"]['type']))

                    if sel.content["enclosures"] and 'href' in sel.content["enclosures"]:
                        if 'type' not in sel.content["enclosures"]:
                            sel.content['enclosures']['type'] = 'unknown'
                        parsed_enclosures.append((sel.content['enclosures']['href'],\
                                    sel.content['enclosures']['type']))

                    if not parsed_enclosures:
                        mainbody += "<br />[ No enclosures. ]<br />"
                    else:
                        for lnk, typ in parsed_enclosures:
                            mainbody += "<a href=\""
                            mainbody += lnk
                            mainbody += "\">["
                            mainbody += typ
                            mainbody += "]</a>\n"

                for attr in list(self.plugin_attrs.keys()):
                    if not attr.startswith("edit_"):
                        continue
                    try:
                        a = getattr(self, attr)
                        (mainbody, extra_content) = a(mainbody, extra_content)
                    except:
                        log.error("Error running Reader edit plugin")
                        log.error(traceback.format_exc())

                # This needn't be prep_for_display'd because the HTML parser
                # handles that.

                content, links = htmlparser.convert(mainbody)

                # 0 always is the mainlink, append other links
                # to the list.

                self.links += links

                if reader_conf['show_description']:
                    s += self.quote_rgx.sub(cc("reader_quote") + "\"\\1\"" + cc.end("reader_quote"), content)

                if reader_conf['enumerate_links']:
                    s += "\n\n"

                    for idx, (t, url, text) in enumerate(self.links):
                        text = prep_for_display(text)
                        url = prep_for_display(url)

                        link_text = "[%B" + str(idx) + "%b][" +\
                                text + "]: " + url + "\n\n"

                        if t == "link":
                            link_text = cc("reader_link") + link_text + cc.end("reader_link")
                        elif t == "image":
                            link_text = cc("reader_image_link") + link_text + cc.end("reader_image_link")

                        s += link_text

        # After we have generated the entirety of the content,
        # strip out any egregious spacing.

        self.text = s.rstrip(" \t\v\n") + extra_content

    def cmd_goto(self, links):
        # link = ( type, url, text )
        hrefs = [ l[1] for l in links ]
        self._goto(hrefs)

    def _toggle_cmd(self, opt):
        c = self.callbacks["get_conf"]()
        c["reader"][opt] = not c["reader"][opt]
        self.callbacks["set_conf"](c)

    def cmd_show_links(self):
        self._toggle_cmd("enumerate_links")

    def cmd_show_desc(self):
        self._toggle_cmd("show_description")

    def cmd_show_encs(self):
        self._toggle_cmd("show_enclosures")

    def cmd_destroy(self):
        self.callbacks["set_var"]("reader_item", None)
        self.callbacks["die"](self)

    def get_opt_name(self):
        return "reader"



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

# The CantoCursesConfig object is a self contained class that it responsible
# for getting a connection to the daemon, getting its configuration settings
# (opts), validating them, and making it (and updates) convenient to use.

# It also contain psuedo-configuration (vars) that aren't actually written to
# the config, but still need to be accessed from various code. These are mostly
# for restoring the screen after a refresh.

# By necessity, it also functions as the thread that watches for added /
# deleted tags, but not for changes to existing tags.

from canto_next.hooks import call_hook
from canto_next.rwlock import RWLock, write_lock, read_lock
from canto_next.remote import assign_to_dict, access_dict

from .locks import config_lock
from .subthread import SubThread

from threading import Thread, Event, current_thread
import traceback
import logging
import curses   # Colors
import json
import re

log = logging.getLogger("CONFIG")

# eval settings need to be somehow converted when read from input.

# These are regexes so that window and color types can be handled with easy
# wildcards, but care has to be taken.

eval_settings = [\
    "defaults\\.rate", "feed\\.rate",
    "defaults\\.keep_time", "feed\\.keep_time",
    "defaults\\.keep_unread", "feed\\.keep_unread",
    "update\\.auto.enabled", "update\\.auto\\.interval",
    "browser\\.text", "taglist\\.border",
    "kill_daemon_on_exit",
    ".*\\.window\\.(maxwidth|maxheight|float)",
    "color\\..*", "tag.(enumerated|collapsed|extra_tags)",
    "reader.(enumerate_links|show_description|show_enclosures)",
    "taglist.(border|tags_enumerated|tags_enumerated_absolute|hide_empty_tags|search_attributes)",
    "taglist.cursor.edge",
    "story.(format_attrs|enumerated)"
]

# Do the one-time compile for the setting regexes. This is called after
# plugins are evaluated, but before curses_start

def finalize_eval_settings():
    global eval_settings
    eval_settings = [ re.compile(x) for x in eval_settings ]

def needs_eval(option):
    for reobj in eval_settings:
        if reobj.match(option):
            return True
    return False

story_needed_attrs = [ "title" ]

CURRENT_CONFIG_VERSION = 1

class CantoCursesConfig(SubThread):

    # The object init just sets up the default settings, doesn't
    # actually do any communication or setup. That's left to init()
    # or, in testing, is ignored.

    def __init__(self):
        self.config_version = 0
        self.vars = {
            "location" : None,
            "error_msg" : "No error.",
            "info_msg" : "No info.",
            "quiet" : False,
            "dispel_msg" : False,
            "input_prompt" : "",
            "input_do_completions" : True,
            "input_completion_root" : None,
            "input_completions" : [],
            "reader_item" : None,
            "reader_offset" : 0,
            "errorbox_offset" : 0,
            "infobox_offset" : 0,
            "selected" : None,
            "target_obj" : None,
            "target_offset" : 0,
            "strtags" : [],
            "curtags" : [],
            "needs_refresh" : False,
            "needs_redraw" : False,
            "needs_resize" : False,
            "transforms" : [],
            "taglist_visible_tags" : [],
        }

        self.validators = {
            "browser" :
            {
                "path" : self.validate_string,
                "text" : self.validate_bool,
            },

            "tags" : self.validate_tags,
            "tagorder" : self.validate_tag_order,

            "update" :
            {
                "style" : self.validate_update_style,
                "auto" :
                {
                    "interval" : self.validate_uint,
                    "enabled" : self.validate_bool,
                }
            },

            "reader" :
            {
                "window" : self.validate_window,
                "key" : self.validate_key,
                "enumerate_links" : self.validate_bool,
                "show_description" : self.validate_bool,
                "show_enclosures" : self.validate_bool,
            },

            "taglist" :
            {
                "window" : self.validate_window,
                "key" : self.validate_key,
                "tags_enumerated" : self.validate_bool,
                "tags_enumerated_absolute" : self.validate_bool,
                "hide_empty_tags" : self.validate_bool,
                "search_attributes" : self.validate_string_list,
                "cursor" : self.validate_taglist_cursor,
                "border" : self.validate_bool,
            },

            "story" :
            {
                "enumerated" : self.validate_bool,
            },

            "input" : { "window" : self.validate_window },

            "errorbox" :
            {
                "window" : self.validate_window,
                "key" : self.validate_key,
            },

            "infobox" :
            {
                "window" : self.validate_window,
                "key" : self.validate_key,
            },

            "main" : { "key" : self.validate_key },

            "screen" : { "key" : self.validate_key },

            "color" : self.validate_color_block,

            "style" : self.validate_style_block,

            "kill_daemon_on_exit" : self.validate_bool
        }

        self.template_config = {
            "config_version" : CURRENT_CONFIG_VERSION,

            "browser" :
            {
                "path" : "xdg-open",
                "text" : False
            },

            "tags" : r"maintag:.*",
            "tagorder" : [],

            "update" :
            {
                "style" : "append",
                "auto" :
                {
                    "interval" : 20,
                    "enabled" : False
                }
            },

            "reader" :
            {
                "window" :
                {
                    "maxwidth" : 0,
                    "maxheight" : 0,
                    "float" : True,
                    "align" : "topleft",
                    "border" : "smart",
                },

                "enumerate_links" : False,
                "show_description" : True,
                "show_enclosures" : True,
                "key" :
                {
                    "space" : "destroy",
                    "s" : "show-summary",
                    "l" : "show-links",
                    "e" : "show-enclosures",
                    "g" : "goto",
                    "down" : "scroll-down",
                    "up" : "scroll-up",
                    "j" : "scroll-down",
                    "k" : "scroll-up",
                    "npage" : "page-down",
                    "ppage" : "page-up",
                    'n' : 'destroy & next-item & item-state read & reader',
                    'p' : 'destroy & prev-item & item-state read & reader',
                    'N' : 'destroy & next-tag & item-state read & reader',
                    'P' : 'destroy & prev-tag & item-state read & reader',
                },
            },

            "taglist" :
            {
                "window" :
                {
                    "maxwidth" : 0,
                    "maxheight" : 0,
                    "float" : False,
                    "align" : "neutral",
                    "border" : "none",
                },

                "tags_enumerated" : False,
                "tags_enumerated_absolute" : False,
                "hide_empty_tags" : True,
                "border" : False,
                "search_attributes" : [ "title" ],

                "key" :
                {
                    "space" : "foritem & item-state read & reader",
                    "g" : "foritems & goto & item-state read & clearitems",
                    "R" : "item-state read *",
                    "U" : "item-state -read *",
                    "r" : "tag-state read",
                    "u" : "tag-state -read",
                    "npage" : "page-down",
                    "ppage" : "page-up",
                    "down" : "next-item",
                    "j" : "next-item",
                    "up" : "prev-item",
                    "k" : "prev-item",
                    "h" : "item-state read",
                    "left" : "item-state read",
                    "l" : "item-state -read",
                    "right" : "item-state -read",
                    "+" : "promote",
                    "-" : "demote",
                    "J" : "next-tag",
                    "K" : "prev-tag",
                    "c" : "toggle-collapse",
                    "C" : "collapse *",
                    "V" : "uncollapse *",
                    "$" : "item-state read tag,0-.",
                    "/" : "search",
                    "n" : "next-marked",
                    "p" : "prev-marked",
                    "M" : "item-state -marked *",
                    "m" : "item-state %marked",
                },

                "cursor" :
                {
                    "type" : "edge",
                    "scroll" : "scroll",
                    "edge" : 5,
                },
            },

            "story" :
            {
                "enumerated" : False,
            },

            "input" :
            {
                "window" :
                {
                    "maxwidth" : 0,
                    "maxheight" : 0,
                    "float" : False,
                    "align" : "bottom",
                    "border" : "none",
                }
            },

            "errorbox" :
            {
                "window" :
                {
                    "maxwidth" : 0,
                    "maxheight" : 0,
                    "float" : True,
                    "align" : "topleft",
                    "border" : "full",
                },

                "key" :
                {
                    "down" : "scroll-down",
                    "up" : "scroll-up",
                    "npage" : "page-down",
                    "ppage" : "page-up",
                    "space" : "destroy",
                }
            },

            "infobox" :
            {
                "window" :
                {
                    "maxwidth" : 0,
                    "maxheight" : 0,
                    "float" : True,
                    "align" : "topleft",
                    "border" : "full",
                },

                "key" :
                {
                    "down" : "scroll-down",
                    "up" : "scroll-up",
                    "j" : "scroll-down",
                    "k" : "scroll-up",
                    "npage" : "page-down",
                    "ppage" : "page-up",
                    "space" : "destroy",
                }
            },

            "main" :
            {
                "key" :
                {
                    ":" : "command",
                    "q" : "quit",
                    "?" : "help",
                    "\\" : "update",
                    "f5" : "update",
                    "C-r" : "refresh",
                }
            },

            "screen" :
            {
                "key" :
                {
                    "tab" : "focus-rel 1",
                }
            },

            "style" :
            {
                "unread" : "%B",
                "read" : "",
                "pending" : "%B",
                "error" : "",
                "marked" : "%B",
                "reader_quote" : "",
                "reader_link" : "",
                "reader_image_link" : "",
                "reader_italics" : "",
                "enum_hints" : "",
                "selected" : "%R",

            },

            "color" :
            {
                "defbg" : -1,
                "deffg" : -1,
                "unread" : 5,
                "read" : 4,
                "pending" : 1,
                "error" : 2,
                "marked" : 8,
                "reader_quote" : 6,
                "reader_link" : 3,
                "reader_image_link" : 5,
                "reader_italics" : 8,
                "enum_hints" : 8,
                "selected" : -1,
            },

            "kill_daemon_on_exit" : False
        }

        for i in range(1, 257):
            self.template_config["color"][str(i)] = i - 1

        self.config = eval(repr(self.template_config))

        self.tag_validators = {
            "enumerated" : self.validate_bool,
            "collapsed" : self.validate_bool,
            "extra_tags" : self.validate_string_list,
            "transform" : self.validate_string,
        }

        self.tag_config = {}

        self.tag_template_config = {
            "enumerated" : False,
            "collapsed" : False,
            "extra_tags" : [],
            "transform" : "None"
        }

        self.daemon_defaults = {}
        self.daemon_feedconf = []

        self.initd = False

    def init(self, backend, compatible_version):
        self.vars["location"] = backend.location_args

        SubThread.init(self, backend)

        self.start_pthread()

        self.version = None
        self.processed = Event()
        self.processed.clear()

        self.write("VERSION", [])
        self.write("WATCHNEWTAGS", [])
        self.write("WATCHDELTAGS", [])
        self.write("LISTTAGS", "")
        self.write("WATCHCONFIGS", "")
        self.write("CONFIGS", [])

        # Spin, may want to convert this into an event, but for now it
        # takes virtually no time and makes it so that we don't have to 
        # check if we're init'd before using *_opt functions.

        while (not self.version):
            pass

        if self.version != compatible_version:
            self.alive = False # Let the subthread die
            return False

        while(not self.initd):
            pass

        self.eval_tags()

        return True

    def validate_uint(self, val, d):
        if type(val) == int and val >= 0:
            return (True, val)
        return (False, False)

    def validate_string(self, val, d):
        if type(val) == str:
            return (True, val)
        return (False, False)

    def validate_bool(self, val, d):
        if val in [ True, False ]:
            return (True, val)
        return (False, False)

    def validate_update_style(self, val, d):
        if val in [ "maintain", "append", "prepend" ]:
            return (True, val)
        return (False, False)

    def validate_tags(self, val, d):
        try:
            re.compile(val)
        except:
            return (False, False)
        return (True, val)

    def validate_tag_order(self, val, d):
        if type(val) != list:
            return (False, False)

        # Strip items no longer relevant
        for item in val[:]:
            if item not in self.vars["strtags"]:
                val.remove(item)

        # Ensure all tags are inluded
        for tag in self.vars["strtags"]:
            if tag not in val:
                val.append(tag)

        return (True, val)

    def validate_window(self, val, d):
        # Ensure all settings exist
        for setting in [ "border", "maxwidth", "maxheight", "align", "float" ]:
            if setting not in val:
                log.debug("Couldn't find %s setting", setting)
                val[setting] = d[setting]

        # Ensure all settings are in their correct range
        if val["border"] not in ["full", "none", "smart"]:
            log.error("border setting must = full OR none OR smart")
            return (False, False)

        if val["float"] not in [ True, False ]:
            log.error("float must be True or False")
            return (False, False)

        for int_setting in ["maxwidth", "maxheight" ]:
            if type(val[int_setting]) != int or val[int_setting] < 0:
                log.error("%s must be a positive integer" % int_setting)
                return (False, False)

        float_aligns = [ "topleft", "topright", "center", "neutral",\
                "bottomleft", "bottomright" ]

        tile_aligns = [ "top", "left", "bottom", "right", "neutral" ]

        if val["float"]:
            if val["align"] not in float_aligns:
                log.error("%s is not a valid alignment for a floating window" % val["align"])
                return (False, False)
        else:
            if val["align"] not in tile_aligns:
                log.error("%s is not a valid alignment for a tiled window" % val["align"])
                return (False, False)

        return (True, val)

    # This doesn't validate that the command will actually work, just that the
    # pair is of the correct types.

    def validate_key(self, val, d):
        if type(val) != dict:
            return (False, False)

        for key in list(val.keys()):
            if type(key) != str:
                return (False, False)

            if type(val[key]) != str:
                return (False, False)

        return (True, val)

    def validate_color(self, val, d, dict_ok=True):
        # Integer, and in the valid color range
        if type(val) == int and val >= -1 and val <= 255:
            return (True, val)

        if type(val) == dict and dict_ok:
            fg_g, bg_g = (False, False)
            r = {}

            # Not specified correctly...
            if "fg" in val:
                fg_g, fg_v = self.validate_color(val["fg"], {}, False)
                if fg_g:
                    r["fg"] = fg_v
            if "bg" in val:
                bg_g, bg_v = self.validate_color(val["bg"], {}, False)
                if bg_g:
                    r["bg"] = bg_v

            if not r:
                return (False, False)
            return (True, r)

        # We have no idea what to do with this crap...
        if type(val) != str:
            return (False, False)

        # See if it's an integer as a string
        try:
            ival = int(val)
            return (True, ival)
        except:
            return color_translate(val)

        return (False, False)

    def migrate_color_block(self, val, d):
        log.warn("Migrating color config to use new color system")
        log.warn("See ':help color' if this butchers your colors")

        r = { "1" : "unread",
              "2" : "read",
              "3" : "reader_link",
              "4" : "reader_image_link",
              "6" : "error",
              "8" : "pending" }

        for key in r:
            if key not in val:
                continue

            # If the given color is a dict, it's fg/bg

            if type(val[key]) == dict:
                if "fg" in val[key] and val[key]["fg"] != -1:
                    val[r[key]] = val[key]["fg"] + 1
                else:
                    log.warn("Ignoring old color %s", key)

            # If the given color is simple, convert it to one of our default
            # pairs

            elif type(val[key]) == int:
                val[r[key]] = val[key] + 1

        # Reset all color pairs to their default

        for i in range(1, 257):
            val[str(i)] = i - 1

        self.write("SETCONFIGS", { "CantoCurses" : {"color" : val }})

    def validate_color_block(self, val, d):
        if type(val) != dict:
            return (False, False)

        r = {}

        for key in val.keys():
            try:
                num = int(key)
                ok, k_v = self.validate_color(val[key], None)
                if ok:
                    r[key] = k_v
                else:
                    log.error("color.%s is invalid (%s)" % (key, val[key]))
                    return (False, False)
                continue
            except:
                pass

            if key in ["deffg", "defbg"]:
                ok, k_v = self.validate_color(val[key], None)
                if ok:
                    r[key] = k_v
                else:
                    log.error("color.%s is invalid (%s)" % (key, val[key]))
                    return (False, False)
            else:
                try:
                    pair = int(val[key])
                    if pair >= -1 and pair <= 255:
                        r[key] = pair
                    else:
                        log.error("color.%s must be >= -1 and <= 255 (%s)" % (key, pair))
                        return (False, False)
                except Exception as e:
                    log.error(e)
                    log.error("color.%s must be an integer, not %s" % (key, val[key]))
                    return (False, False)

        if self.config_version < 1:
            self.migrate_color_block(r, d)

        for key in d.keys():
            if key not in r:
                r[key] = d[key]

        return (True, r)

    # We don't care if the styles actually make sense, only that they won't
    # cause trouble when used as a string.

    def validate_style_block(self, val, d):
        if type(val) != dict:
            return (False, False)

        r = eval(repr(d))

        for key in val:
            if type(key) != str or type(val[key]) != str:
                log.error("Ignoring style %s - %s", key, val[key])
            else:
                r[key] = val[key]

        return (True, r)

    def validate_string_list(self, val, d):
        if type(val) != list:
            return (False, False)

        r = []
        for item in val:
            if type(item) == str:
                r.append(item)
            else:
                return (False, False)

        return (True, r)

    def validate_taglist_cursor(self, val, d):
        if type(val) != dict:
            return (False, False)

        for setting in [ "type","scroll","edge" ]:
            if setting not in val:
                val[setting] = d[setting]

        if val["type"] not in ["edge","top","middle","bottom"]:
            log.error("Cursor type %s unknown!" % (val["type"],))
            return (False, False)

        if val["scroll"] not in ["scroll", "page"]:
            log.error("Cursor scroll type %s unknown!" % (val["scroll"],))

        if type(val["edge"]) != int or val["edge"] < 0:
            log.error("Cursor edge invalid, must be int >= 0: %s" % (val["edge"],))
            return (False, False)

        return (True, val)

    def _list_diff(self, cur, old):
        adds = []
        dels = []

        for item in old:
            if item not in cur:
                dels.append(item)

        for item in cur:
            if item not in old:
                adds.append(item)

        return (adds, dels)

    # Recursively validate config c, with validators in v, falling back on d
    # when it failed. Return a dict containing all of the changes actually
    # made.

    # Note that unknown values are detected only to avoid access errors, they
    # are totally ignored and will never get changes processed.

    def validate_config(self, c, d, v):
        changes = {}
        deletions = {}

        # Sub in non-existent values:

        for key in list(v.keys()):
            if key not in c:
                c[key] = d[key]

        # Validate existing values.

        for key in list(c.keys()):
            # Unknown values, don't validate
            if key not in v:
                continue

            # Key is section, recurse, only add changes if there
            # are actual changes.

            elif type(v[key]) == dict:
                chgs, dels = self.validate_config(c[key], d[key], v[key])
                if chgs:
                    changes[key] = chgs
                if dels:
                    deletions[key] = dels

            # Key is basic, validate
            else:
                good, val = v[key](c[key], d[key])

                # Value is good, pass on
                if good:
                    if val != d[key]:
                        dels = {}
                        if type(val) == list:
                            chgs, dels, = self._list_diff(val, d[key])
                        elif type(val) == dict and type(d[key]) == dict:
                            for d_key in d[key].keys():
                                if d_key not in c[key].keys():
                                    dels[d_key] = "DELETE"
                        if dels:
                            deletions[key] = dels
                        changes[key] = val
                    c[key] = val

                # Value is bad, revert
                else:
                    err = "config %s was bad (%s) reverting to default (%s)" %\
                            (key, c[key], d[key])

                    #if not self.glog_handler:
                    #    self.early_errors.append(err)
                    #    log.error("Will display " + err)
                    #else:
                    log.error(err)

                    changes[key] = d[key]
                    c[key] = d[key]

        for key in list(d.keys()):
            if key not in c and key not in v:
                deletions[key] = "DELETE"

        return changes, deletions

    # We use strtags to validate tag order, and also to populate the
    # TagUpdater()

    @write_lock(config_lock)
    def prot_listtags(self, tags):
        self.vars["strtags"] = tags
        self.config["tagorder"] = tags

    def prot_version(self, version):
        self.version = version

    def prot_pong(self, empty):
        self.processed.set()

    def wait_write(self, cmd, args):
        self.write(cmd, args)
        if current_thread() != self.prot_thread:
            self.write("PING", [])
            self.processed.wait()
            self.processed.clear()

    # configs accepts any changes, calls the opt_change hooks and if write is
    # set, sends those changes to the daemon. It's called both when receving
    # CONFIGS from the daemon and when we change opts internally (thus the
    # write flag).

    # Note that changes are the only ones propagated through hooks because they
    # are a superset of deletions (i.e. a deletion counts as a change).

    @write_lock(config_lock)
    def prot_configs(self, given, write = False):
        log.debug("prot_configs given:\n%s\n", json.dumps(given, indent=4, sort_keys=True))
        if "tags" in given:
            for tag in list(given["tags"].keys()):
                ntc = given["tags"][tag]

                tc = self.get_tag_conf(tag)

                changes, deletions =\
                        self.validate_config(ntc, tc, self.tag_validators)

                if write:
                    if changes:
                        self.wait_write("SETCONFIGS", { "tags" : { tag : changes }})
                    if deletions:
                        self.wait_write("DELCONFIGS", { "tags" : { tag : deletions }})

                if changes:
                    self.tag_config[tag] = ntc
                    call_hook("curses_tag_opt_change", [ { tag : changes } ])

        if "CantoCurses" in given:
            new_config = given["CantoCurses"]

            if "config_version" in new_config:
                self.config_version = new_config["config_version"]

            changes, deletions =\
                    self.validate_config(new_config, self.config,\
                    self.validators)

            if "config_version" not in new_config or\
                    new_config["config_version"] != CURRENT_CONFIG_VERSION:

                log.debug("Configuration migrated from %s to %s",\
                        self.config_version, CURRENT_CONFIG_VERSION)

                self.config_version = CURRENT_CONFIG_VERSION
                new_config["config_version"] = CURRENT_CONFIG_VERSION
                changes["config_version"] = CURRENT_CONFIG_VERSION
                self.write("SETCONFIGS", { "CantoCurses" : {"config_version" : CURRENT_CONFIG_VERSION } })

            if write:
                if changes:
                    self.wait_write("SETCONFIGS", { "CantoCurses" : changes })

                if deletions:
                    self.wait_write("DELCONFIGS", { "CantoCurses" : deletions })

            if changes:
                self.config = new_config
                call_hook("curses_opt_change", [ changes ])
                if "tags" in changes:
                    self.eval_tags()

        if "defaults" in given:

            changes = {}

            for key in given["defaults"]:
                if key in self.daemon_defaults:
                    if given["defaults"][key] != self.daemon_defaults[key]:
                        changes[key] = given["defaults"][key]
                else:
                    changes[key] = given["defaults"][key]

            self.daemon_defaults.update(changes)

            if write:
                self.wait_write("SETCONFIGS", { "defaults" : self.daemon_defaults })

            call_hook("curses_def_opt_change", [ changes ])

        if "feeds" in given:

            self.daemon_feedconf = given["feeds"]
            if write:
                self.wait_write("SETCONFIGS", { "feeds" : self.daemon_feedconf })

            call_hook("curses_feed_opt_change", [ given["feeds"] ])

        self.initd = True

    # Process new tags.

    @write_lock(config_lock)
    def prot_newtags(self, tags):

        if not self.initd:
            for tag in tags:
                if tag not in self.vars["strtags"]:
                    self.vars["strtags"].append(tag)
                if tag not in self.config["tagorder"]:
                    self.config["tagorder"].append(tag)
            return

        c = self.get_conf()

        # Likely the same as tags
        changes = False
        newtags = []

        for tag in tags:
            if tag not in c["tagorder"]:
                c["tagorder"] = c["tagorder"] + [ tag ]
                changes = True

            if tag not in self.vars["strtags"]:

                # If we don't have configuration for this
                # tag already, substitute the default template.

                if tag not in self.tag_config:
                    log.debug("Using default tag config for %s", tag)
                    self.tag_config[tag] = self.tag_template_config.copy()

                self.vars["strtags"].append(tag)
                newtags.append(tag)
                changes = True

        # If there aren't really any tags we didn't know about, no bail.

        if not changes:
            return

        self.set_conf(c)

        for tag in newtags:
            log.debug("New tag %s", tag)
            call_hook("curses_new_tag", [ tag ])

        self.eval_tags()

    @write_lock(config_lock)
    def prot_deltags(self, tags):
        if not self.initd:
            for tag in tags:
                if tag in self.vars["strtags"]:
                    self.vars["strtags"].remove(tag)
                if tag in self.config["tagorder"]:
                    self.config["tagorder"].append(tag)
            return

        c = self.get_conf()
        changes = False

        for tag in tags:
            if tag in self.vars["strtags"]:
                if tag in c["tagorder"]:
                    c["tagorder"] = [ x for x in self.config["tagorder"] if x != tag ]
                    changes = True
                self.vars["strtags"].remove(tag)
                call_hook("curses_del_tag", [ tag ])
                self.eval_tags()
            else:
                log.debug("Got DELTAG for non-existent tag!")

        if changes:
            self.set_conf(c)

    @write_lock(config_lock)
    def eval_tags(self):
        prevtags = self.vars["curtags"]

        sorted_tags = []
        r = re.compile(self.config["tags"])

        for tag in self.vars["strtags"]:

            # This can happen between the time that a tag is removed from the config
            # and the time that we receive a DELTAG event.
            if tag not in self.config["tagorder"]:
                continue

            elif r.match(tag):
                sorted_tags.append((self.config["tagorder"].index(tag), tag))
        sorted_tags.sort()

        self.set_var("curtags", [ x for (i, x) in sorted_tags ])

        if not self.vars["curtags"]:
            log.warn("NOTE: Current 'tags' setting eliminated all tags!")

        # If evaluated tags differ, we need to let other know.

        if prevtags != self.vars["curtags"]:
            log.debug("Evaluated Tags Changed:\n%s\n", json.dumps(self.vars["curtags"], indent=4))
            call_hook("curses_eval_tags_changed", [])

    def set_var(self, tweak, value):
        # We only care if the value is different, or it's a message
        # value, which should always cause a fresh message display,
        # even if it's the same error as before.

        if self.vars[tweak] != value:
            self.vars[tweak] = value
            call_hook("curses_var_change", [{ tweak : value }])

    def get_var(self, tweak):
        if tweak in self.vars:
            return self.vars[tweak]
        raise Exception("Unknown variable: %s" % (tweak,))

    # Overall configuration operation functions. The paradigm is that internal
    # code can "get" the conf, which is a copy of the real conf, modify it,
    # then "set" the conf which will properly process the changes.

    # prot_configs handles locking

    def set_conf(self, conf):
        self.prot_configs({"CantoCurses" : conf }, True)

    def set_tag_conf(self, tag, conf):
        self.prot_configs({ "tags" : { tag : conf } }, True)

    def set_def_conf(self, conf):
        self.prot_configs({ "defaults" : conf }, True)

    def set_feed_conf(self, name, conf):
        config_lock.acquire_read()
        d_f = eval(repr(self.daemon_feedconf), {}, {})
        config_lock.release_read()

        for f in d_f:
            if f["name"] == name:
                log.debug("updating %s with %s", f, conf)
                f.update(conf)
                break
        else:
            d_f.append(conf)

        self.prot_configs({ "feeds" : d_f }, True)

    @read_lock(config_lock)
    def get_conf(self):
        return eval(repr(self.config), {}, {})

    @read_lock(config_lock)
    def get_tag_conf(self, tag):
        if tag in self.tag_config:
            return eval(repr(self.tag_config[tag]), {}, {})
        return eval(repr(self.tag_template_config), {}, {})

    @read_lock(config_lock)
    def get_def_conf(self):
        return eval(repr(self.daemon_defaults), {}, {})

    @read_lock(config_lock)
    def get_feed_conf(self, name):
        for f in self.daemon_feedconf:
            if f["name"] == name:
                return eval(repr(f), {}, {})
        return None

    @write_lock(config_lock)
    def set_opt(self, option, value):
        c = self.get_conf()
        assign_to_dict(c, option, value)
        self.set_conf(c)

    @read_lock(config_lock)
    def get_opt(self, option):
        c = self.get_conf()
        valid, value = access_dict(c, option)
        if not valid:
            return None
        return value

    @write_lock(config_lock)
    def set_tag_opt(self, tag, option, value):
        tc = self.get_tag_conf(tag)
        assign_to_dict(tc, option, value)
        self.set_tag_conf(tag, tc)

    @read_lock(config_lock)
    def get_tag_opt(self, tag, option):
        tc = self.get_tag_conf(tag)
        valid, value = access_dict(tc, option)
        if not valid:
            return None
        return value

    @write_lock(config_lock)
    def switch_tags(self, tag1, tag2):
        c = self.get_conf()

        t1_idx = c["tagorder"].index(tag1)
        t2_idx = c["tagorder"].index(tag2)

        c["tagorder"][t1_idx] = tag2
        c["tagorder"][t2_idx] = tag1

        self.set_conf(c)

        self.eval_tags()

config = CantoCursesConfig()



# Smart Link Plugin
# by Jack Miller
# v1.0

# Allow links to be fetched to disk and then run a smarter handler on it

# Use :fetch instead of :goto to use these handlers.

# HANDLERS is a list of handlers, specified like:

HANDLERS = [
    { "match-url" : ".*\\.mp3$",
      "handler" : "mplayer",
      "regex" : True,
    },
    { "match-file" : "image data",
      "handler" : "feh",
    },
    { "match-file" : "PDF",
      "handler" : "evince",
    }
]

# Each handler will either have a 'match-url' setting or a 'match-file' setting.

# These will be considered regexes if 'regex' is True, it defaults to False,
# which will use a basic string search.

# 'match-url' should be used for network capable helpers, like mplayer, and if
# a match-url match is found, canto-curses won't download the file.

# If nothing matches ANY 'match-url' handler, then the file is downloaded and
# `file` is run on it. 'match-file' handlers are then checked against that
# output.

# If no handlers match, then your default browser is used, as if you had used
# :goto instead of :fetch.

# Matches are searched in order (except url matches always come before file
# matches). So if you want to, say, open PNGs in one sort of handler, and all
# other images in another handler, you could list something like

# { "match-file" : "PNG image data",
#   "handler" : "png-helper",
# },
# { "match-file" : "image data",
#   "handler" : "other-helper",
# }

# You may also specify a "pause" setting if you want to use other terminal
# programs without canto getting in the way.

# { "match-file" : "Audio data",
#   "handler" : "mocp",
#   "pause" : True,
# }

# If you have no match-file handlers listed, the file will not be downloaded.

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_next.hooks import on_hook

from canto_curses.taglist import TagListPlugin
from canto_curses.reader import ReaderPlugin
from canto_curses.command import register_commands

from threading import Thread
import subprocess
import tempfile
import logging
import urllib
import shlex
import sys
import re
import os

log = logging.getLogger("SMARTLINK")

class SmartLinkThread(Thread):
    def __init__(self, base_obj, href):
        Thread.__init__(self, name="Smart Link: %s" % href)
        self.base_obj = base_obj
        self.href = href
        self.daemon = True

    def run(self):
        got_handler = None
        file_handlers = []

        for handler in HANDLERS:
            # If there's no handler defined, we don't care if it's a match
            if "handler" not in handler:
                log.error("No handler binary defined for: %s" % handler)
                continue

            if "regex" in handler and handler["regex"] == True:
                handler['regex'] = True
            else:
                handler['regex'] = False

            if "match-url" in handler:
                got_handler = self.try_handler(handler, "url", self.href)
                if got_handler:
                    break
            elif "match-file" in handler:
                file_handlers.append(handler)
            else:
                log.error("No match-url or match-file in handler %s" % handler)
        else:
            # We didn't find a matching URL handler.

            # No file_handlers present, don't bother to download, create
            # a default browser handler.

            if file_handlers:
                try:
                    tmpnam = self.grab_it()
                except Exception as e:
                    log.error("Couldn't download file: %s" % e)

                    # If we couldn't get the file, skip all of these
                    file_handlers = []
                else:
                    try:
                        fileoutput = subprocess.check_output("file %s" % shlex.quote(tmpnam), shell=True)
                        fileoutput = fileoutput.decode()
                    except Exception as e:
                        log.error("Couldn't get file output: %s" % e)

                        # If we couldn't get the `file` output, also skip
                        file_handlers = []

            for f_handler in file_handlers:
                log.debug("f_handler: %s", f_handler)
                got_handler = self.try_handler(f_handler, "file", fileoutput)
                if got_handler:
                    self.href = tmpnam
                    break
            else:
                conf = self.base_obj.callbacks["get_conf"]()
                got_handler = { 
                        "handler" : conf["browser"]["path"],
                        "text" : conf["browser"]["text"]
                }

        # Okay, so at this point we have self.href, which is either the URL or
        # the temporary file path, and got_handler telling us what to invoke an
        # how.

        log.info("Opening %s with %s" % (self.href, got_handler["handler"]))

        # Make sure that we quote href such that malicious URLs like
        # "http://example.com & rm -rf ~/" won't be interpreted by the shell.

        href = shlex.quote(self.href)

        pause = False
        if "pause" in got_handler and got_handler["pause"]:
            self.base_obj.callbacks["pause_interface"]()
            pause = True

        path = got_handler["handler"]
        if "%u" in path:
            path = path.replace("%u", href)
        elif href:
            path = path + " " + href

        pid = os.fork()

        if not pid:
            # A lot of programs don't appreciate having their fds closed, so
            # instead we dup them to /dev/null.

            fd = os.open("/dev/null", os.O_RDWR)
            os.dup2(fd, sys.stderr.fileno())

            if not pause:
                os.setpgid(os.getpid(), os.getpid())
                os.dup2(fd, sys.stdout.fileno())
                os.dup2(fd, sys.stdin.fileno())

            os.execv("/bin/sh", ["/bin/sh", "-c", path])

            # Just in case.
            sys.exit(0)

        # Parent process only cares if we should wait for the process to finish

        elif pause:
            os.waitpid(pid, 0)
            self.base_obj.callbacks["unpause_interface"]()

    def try_handler(self, handler, suffix, content):
        got_match = False
        element = "match-" + suffix

        # We know these elements exist, from above
        if not handler["regex"]:
            got_match = handler[element] in content
        else:
            try:
                got_match = re.match(handler[element], content)
            except Exception as e:
                log.error("Failed to do %s match: $s" % (suffix, e))

        if got_match:
            return handler

    def grab_it(self):
        # Prepare temporary files

        # Get a base filename (sans query strings, etc.) from the URL

        tmppath = urllib.parse.urlparse(self.href).path
        fname = os.path.basename(tmppath)

        # Grab a temporary directory. This allows us to create a file with
        # an unperturbed filename so scripts can freely use regex /
        # extension matching in addition to mimetype detection.

        tmpdir = tempfile.mkdtemp(prefix="canto-")
        tmpnam = tmpdir + '/' + fname

        log.debug("Downloading %s to %s", self.href, tmpnam)

        on_hook("curses_exit", lambda : (os.unlink(tmpnam)))
        on_hook("curses_exit", lambda : (os.rmdir(tmpdir)))

        tmp = open(tmpnam, 'w+b')

        # Set these because some sites think python's urllib is a scraper and
        # 403 it.

        extra_headers = { 'User-Agent' :\
                'Canto/0.9.0 + http://codezen.org/canto-ng'}

        request = urllib.request.Request(self.href, headers = extra_headers)

        # Grab the HTTP info / prepare to read.
        response = urllib.request.urlopen(request)

        # Grab in kilobyte chunks to avoid wasting memory on something
        # that's going to be immediately written to disk.

        while True:
            r = response.read(1024)
            if not r:
                break
            tmp.write(r)

        response.close()
        tmp.close()

        return tmpnam

class TagListSmartLink(TagListPlugin):
    def __init__(self, taglist):
        self.taglist = taglist
        self.plugin_attrs = {}

        cmds = {
            "fetch" : (self.cmd_fetch_link, ["item-list"], "Fetch link"),
        }

        register_commands(taglist, cmds)

        taglist.bind('f', 'fetch')

    def cmd_fetch_link(self, items):
        SmartLinkThread(self.taglist, items[0].content["link"]).start()
        
class ReaderSmartLink(ReaderPlugin):
    def __init__(self, reader):
        self.reader = reader
        self.plugin_attrs = {}

        cmds = {
            "fetch" : (self.cmd_fetch_link, ["link-list"], "Fetch link"),
        }

        register_commands(reader, cmds)

        reader.bind('f', 'fetch')

    def cmd_fetch_link(self, link):
        SmartLinkThread(self.reader, link[0][1]).start()



# Yank Plugin
# by Jack Miller
# v1.1

# Requires xclip to be somewhere in $PATH

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_curses.taglist import TagListPlugin
from canto_curses.command import register_commands

from subprocess import Popen, PIPE
from os import system
import logging

log = logging.getLogger("YANK")

def yank(content):
    p = Popen(["xclip"], stdin=PIPE)
    p.communicate(content.encode("utf-8"))

class TagListYank(TagListPlugin):
    def __init__(self, taglist):
        self.plugin_attrs = {}

        cmds = {
            "yank-link" : (self.cmd_yank_link, ["item-list"], "Yank link"),
            "yank-title" : (self.cmd_yank_title, ["item-list"], "Yank title"),
        }
        register_commands(taglist, cmds)

        taglist.bind('y', 'yank-link')
        taglist.bind('Y', 'yank-title')

    def cmd_yank_link(self, items):
        yank(items[0].content["link"])

    def cmd_yank_title(self, items):
        yank(items[0].content["title"])



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import PluginHandler, Plugin

from .tagcore import tag_updater

import traceback
import logging
import curses
import shlex

import readline

log = logging.getLogger("COMMAND")

cmds = {}
arg_types = {}
aliases = {}

# These objects don't really need to be objects, but they're cleaner than
# tossing around a zillion tuples.

class CantoCommand(object):
    def __init__(self, obj, name, func, args, help_txt, group):
        self.name = name
        self.obj = obj              # 0
        self.func = func            # 1
        self.args = args            # 2
        self.help_txt = help_txt    # 3
        self.group = group          # 4

class CantoArgType():
    def __init__(self, obj, name, help_txt, validator, hook):
        self.name = name
        self.obj = obj                  # 0
        self.help_txt = help_txt        # 1
        self.validator = validator      # 2
        self.hook = hook                # 3

class CantoAlias():
    def __init__(self, obj, alias, longform):
        self.alias = alias
        self.obj = obj              # 0
        self.longform = longform    # 1

def register_command(obj, name, func, args, help_txt, group="hidden"):
    c = CantoCommand(obj, name, func, args, help_txt, group)
    if name not in cmds:
        cmds[name] = [ c ]
    else:
        cmds[name].append(c)

def register_commands(obj, cmds, group="hidden"):
    for name in cmds:
        func, args, help_text = cmds[name]
        register_command(obj, name, func, args, help_text, group)

def commands(g=None):
    c = []

    for ck in cmds.keys():
        group = cmds[ck][-1].group
        if group == "hidden":
            continue
        if g and g != group:
            continue

        for ak in aliases.keys():
            if aliases[ak][-1].longform == ck:
                c.append(ak)
                break
        else:
                c.append(ck)
    return c

def groups():
    g = []
    for ck in cmds.keys():
        grp = cmds[ck][-1].group
        if grp != "hidden" and grp not in g:
            g.append(grp)
    return g

def command_help(command, detailed=False):
    lookup = shlex.split(command)
    working_cmd = _get_max_sig(lookup)[-1]

    if not detailed:
        s = "%s" % working_cmd.help_txt
        if '\n' in s:
            s = s[:s.index('\n')]
    else:
        s = "%s %s\n" % (command, " ".join(["[" + x + "]" for x in working_cmd.args]))
        s += "\n%s" % working_cmd.help_txt
        for arg in working_cmd.args:
            s += "\n\n"
            s += arg_types[arg][-1].help_txt
    return s

def register_arg_type(obj, name, help_txt, validator, hook=None):
    at = CantoArgType(obj, name, help_txt, validator, hook)
    if name not in arg_types:
        arg_types[name] = [ at ]
    else:
        arg_types[name].append(at)

def register_arg_types(obj, types):
    for name in types:
        register_arg_type(obj, name, *types[name])

def register_alias(obj, alias, longform):
    a = CantoAlias(obj, alias, longform)
    if alias in aliases:
        aliases[alias].append(a)
    else:
        aliases[alias] = [ a ]

def register_aliases(obj, given):
    for alias in given:
        register_alias(obj, alias, given[alias])

# Passthru for any string, including empty
def _string():
    return (None, lambda x : (True, x))

def word():
    def word_validator(x):
        for c in ' \t':
            if c in x:
                return (False, x)
        return (True, x)
    return (None, word_validator)

register_arg_type(_string, "string", "[string] Any string", _string)
register_arg_type(word, "word", "[word] Any word (no whitespace)", word)

# Unregister, clear out obj associations, del keys if empty.

def _unregister(obj, dct, name):
    if name in dct:
        dct[name] = [ x for x in dct[name] if x.obj != obj]
        if not dct[name]:
            del dct[name]

def unregister_command(obj, name):
    _unregister(obj, cmds, name)

def unregister_arg_type(obj, typ):
    _unregister(obj, arg_types, typ)

def unregister_alias(obj, alias):
    _unregister(obj, aliases, alias)

def unregister_all(obj):
    for key in list(cmds.keys()):
        unregister_command(obj, key)
    for key in list(arg_types.keys()):
        unregister_arg_type(obj, key)
    for key in list(aliases.keys()):
        unregister_alias(obj, key)

# Take a split lookup and unalias the first argument

def _unalias(lookup):
    longest_alias = ""

    # Re-combine to match across multiple tokens
    total = " ".join([ shlex.quote(x) for x in lookup])

    # Commands are automatically aliases of themselves, so that, for example
    # "quit" won't be expanded into "quituit"

    possibles = list(aliases.keys())
    possibles.extend(cmds.keys())

    # Expand an alias into the lookup
    for alias in possibles:
        if not total.startswith(alias):
            continue
        if len(alias) > len(longest_alias):
            longest_alias = alias

    if longest_alias == "" or longest_alias in cmds:
        return lookup

    # deref -1 for latest register, 1 for longform instead of obj
    total = total.replace(longest_alias, aliases[longest_alias][-1].longform, 1)

    log.debug("Unaliased to: %s", total)

    return shlex.split(total)

# Use lookup information to find longest possible sig So, given
# ['remote','addfeed'], return the signature for "remote addfeed" instead of
# just "remote". This lets us get completions for specific subcommands.

# Returns the a tuple with sig info, and a match, which has the command
# stripped out.

def _get_max_sig(lookup):
    lookup = _unalias(lookup)
    match = []
    ret = None

    for i in range(len(lookup)):
        test = " ".join(lookup[0:i + 1])

        if test in cmds:
            ret = cmds[test][-1]
            match = lookup[i + 1:]
        else:
            break

    return match, ret

def cmd_complete_info():
    buf = readline.get_line_buffer()

    lookup = shlex.split(buf)

    # If there's a space, we've moved on to a new argument, so stub in an empty
    # partial argument.

    if not buf or buf[-1] == ' ':
        lookup.append('')
        prefix = ''
    else:
        prefix = lookup[-1]

    if len(lookup) == 1:
        c = list(cmds.keys())
        c.extend(list(aliases.keys()))
        c.sort()
        log.debug("CMDS: %s", c)
        return ("", "", c)
    else:
        lookup, sig = _get_max_sig(lookup)

        # No matches, bail
        if not sig:
            return None

        # No completing beyond end of arguments

        if len(lookup) > len(sig.args):
            log.debug("completing too many args")
            return None

        # XXX these should check that type exists for plugins

        # validate that the arguments we're not completing are okay
        # so that we don't tab complete a broken command.

        for i, typ in enumerate(sig.args[:len(lookup) - 1]):
            completions, validator = arg_types[typ][-1].validator()
            if not validator(lookup[i]):
                return None

        # now get completions for the actual terminating command

        at = arg_types[sig.args[len(lookup) - 1]][-1]
        if at.hook:
            at.hook()
        completions, validator = at.validator()
        return (sig.help_txt, at.help_txt, completions)
    return None

def cmd_execute(cmd):
    lookup = shlex.split(cmd)

    if not lookup:
        return False

    lookup, sig = _get_max_sig(lookup)

    if not sig:
        return False

    args = []

    for i, typ in enumerate(sig.args):
        completions, validator = arg_types[typ][-1].validator()

        # If we're on the last part of the sig, and there's more than one
        # argument remaining, then smash them together in such a way that
        # shlex.split will properly reparse them.

        if i == len(sig.args) - 1 and len(lookup) > (i + 1):
            token = " ".join([ shlex.quote(x) for x in lookup[i:]])
        elif i < len(lookup):
            token = lookup[i]
        else:
            token = ""

        okay, r = validator(token)

        if not okay:
            log.info("'%s' is not a valid %s" % (token, typ))
            return False
        args.append(r)

    sig.func(*args)
    return True

# Return a function taking a string definition of a list, with possible special
# characters, and return an explicit list. Each item in the returned list will be unique, so
# doing something like 1,2,* won't repeat items 1 and 2.

# name - type name (for decent error output)
# itrs - a dict of iterables being indexed (i.e. a lists of items), each key is a domain
# syms - symbolics (i.e. { 'domain' : { '*' : all_items, '.' : [ current_item ]}})
# fallback - list of items to return if no indices
# s - input string to parse

# This is likely used lambda x: _int_range("mytype", {}, {}, x) to encapsulate the rest
# of the state from the command infrastructure

def _range(cur_iter, syms, item):
    cur_syms = syms[cur_iter]

    # Convert into indices
    if item in cur_syms:

        # This will default to the first item if passed a sym with more
        # items (*)

        return cur_syms[item][0]
    else:
        try:
            item = int(item)
            return item
        except:
            pass
    return None

def _int_check(x):
    try:
        r = int(x)
        return (True, r)
    except:
        return (False, None)

def _int_range(name, itrs, syms, fallback, s):
    s.strip()

    if (not s):
        if fallback:
            log.debug("falling back to %s", fallback)
            return (True, fallback)
        return (True, [])

    slist = s.split(',')

    # Default domain is 'all'
    cur_iter = 'all'

    idxlist = []

    # Convert slist into a list of numeric indices

    for item in slist:
        item.strip()

        # Deal with ranges
        if "-" in item:
            start, stop = item.split('-')

            start_idx = _range(cur_iter, syms, start)
            stop_idx = _range(cur_iter, syms, stop)

            if start_idx == None:
                log.warn("Couldn't convert range start: %s" % start)
                continue
            if start_idx < 0 or start_idx >= len(itrs[cur_iter]):
                log.warn("Range start out of bounds: %s (%s)" % (start_idx, len(itrs[cur_iter])))
                continue
            if stop_idx == None:
                log.warn("Couldn't convert range stop: %s" % stop)
                continue
            if stop_idx < 0 or stop_idx >= len(itrs[cur_iter]):
                log.warn("Range stop out of bounds: %s (%s)" % (stop_idx, len(itrs[cur_iter])))
                continue

            idxlist.extend([ (cur_iter, x) for x in range(start_idx,stop_idx + 1) ])

        # Convert specials... note that domains come before syms, but it would
        # be a bad idea to have conflicts anyway.

        elif item in itrs:
            cur_iter = item
        elif item in syms[cur_iter]:
            idxlist.extend([ (cur_iter, x) for x in syms[cur_iter][item] ])
        else:
            try:
                r = int(item)
                idxlist.append((cur_iter, r))
            except:
                if item or ( item == '' and not fallback ):
                    log.warn("Invalid %s : '%s'" % (name, item))

    # Strip down to unique indices

    uidxlist = []
    for tup in idxlist:
        if tup not in uidxlist:
            uidxlist.append(tup)

    # Convert into list of items in itr

    rlist = []
    for domain, idx in uidxlist:
        if 0 <= idx < len(itrs[domain]):
            if itrs[domain][idx] not in rlist:
                rlist.append(itrs[domain][idx])
        else:
            log.warn("%s out of range of %s domain: %s idx with len %s" % (name, domain, idx, len(itrs[domain])))

    return (True, rlist)

class CommandPlugin(Plugin):
    pass

class CommandHandler(PluginHandler):
    def __init__(self):
        PluginHandler.__init__(self)

        self.plugin_class = CommandPlugin
        self.update_plugin_lookups()

        self.key_translations =\
                { '.' : "period",
                  '\t' : "tab",
                  "C-i" : "tab",
                  ' ' : "space",
                  "\\" : "\\\\" }

        self.meta = False

    def translate_key(self, key):
        if key in self.key_translations:
            return self.key_translations[key]
        return key

    def key(self, k):

        # Translate numeric key into config friendly keyname

        optname = self.get_opt_name() + ".key."

        # Add meta prefix.
        if self.meta:
            if k >= 64:
                k -= 64
                optname += "M-"
            self.meta = False

        if k > 255:
            for attr in dir(curses):
                if not attr.startswith("KEY_"):
                    continue

                if k == getattr(curses, attr):
                    optname += attr[4:].lower()

        # Remember meta for next keypress.
        elif curses.ascii.ismeta(k):
            self.meta = True
            return None
        else:
            keyname = ""
            # Add ctrl prefix.
            if curses.ascii.iscntrl(k):
                keyname += "C-"
                k += 96

            keyname += chr(k)
            optname += self.translate_key(keyname)

        log.debug("trying key: %s", optname)

        try:
            r = self.callbacks["get_opt"](optname)
        except:
            r = None

        # None happens if the option is unset
        # "None" can be used by the user to ignore
        # a keybind without any chatter.
        if r and r != "None":
            return r

        return None




# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.hooks import on_hook

from .config import config

# color code, convert a symbolic color name (e.g. "unread") into a code to put
# in a theme string

class CantoColorManager:
    def __init__(self):
        self.color_conf = config.get_opt("color")
        self.style_conf = config.get_opt("style")
        on_hook("curses_opt_change", self.on_opt_change, self)

    def on_opt_change(self, config):
        if "color" in config:
            self.color_conf = config["color"]
        if "style" in config:
            self.style_conf = config["style"]

    def _invert(self, codes):
        inverted = ""

        in_code = False
        long_code = False

        for c in codes:
            if in_code:
                if long_code:
                    if c == "]":
                        inverted += "%0"
                        long_code = False
                else:
                    if c in "12345678":
                        inverted += "%0"
                    elif c in "BRDSU":
                        inverted += "%" + c.lower()
                    elif c == "[":
                        long_code = True
                    in_code = False
            elif c == "%":
                in_code = True
        return inverted

    def __call__(self, name):
        color = ""

        if self.color_conf[name] > 8:
            color = "%[" + str(self.color_conf[name]) + "]"
        elif self.color_conf[name] > 0:
            color = "%" + str(self.color_conf[name])

        return color + self.style_conf[name]

    def end(self, name):
        return self._invert(self(name))

cc = CantoColorManager()



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import Plugin
from .guibase import GuiBase
from .widecurse import get_rlpoint
from .command import cmd_complete_info

import logging
log = logging.getLogger("INPUT")

import readline
import curses
import shlex
from curses import ascii

class InputPlugin(Plugin):
    pass

class InputBox(GuiBase):
    def __init__(self):
        GuiBase.__init__(self)
        self.plugin_class = InputPlugin
        self.pad = None

    def init(self, pad, callbacks):
        self.pad = pad

        self.callbacks = callbacks

        self.reset()

    def reset(self):
        self.pad.erase()
        self.pad.addstr(self.callbacks["get_var"]("input_prompt"))
        self.minx = self.pad.getyx()[1]
        self.x = self.minx

        # Part that's not considered
        self.completion_root = None
        self.completions = None

    def _get_prefix(self):
        buf = readline.get_line_buffer()
        if (not buf) or buf[-1].isspace():
            prefix = ""
        else:
            prefix = shlex.split(buf)[-1]
        return prefix

    def rotate_completions(self):
        complist = self.callbacks["get_var"]("input_completions")
        oldpref = self.callbacks["get_var"]("input_completion_root")

        prefix = self._get_prefix()
        if not complist or oldpref != prefix:
            r = cmd_complete_info()
            if not r or not r[2]:
                complist = []
            else:
                complist = [ x[len(prefix):] for x in r[2] if x.startswith(prefix) ]
                complist.sort()
        else:
            complist = complist[1:] + [ complist[0] ]

        self.callbacks["set_var"]("input_completions", complist)
        self.callbacks["set_var"]("input_completion_root", prefix)

    def break_completion(self):
        comp = self.callbacks["get_var"]("input_completions")
        self.callbacks["set_var"]("input_completions", [])
        self.callbacks["set_var"]("input_completion_root", "")
        if comp:
            return comp[0]
        return None

    def refresh(self):
        if not self.pad:
            return

        self.pad.move(0, self.minx)
        maxx = self.pad.getmaxyx()[1]

        s = readline.get_line_buffer()
        complist = self.callbacks["get_var"]("input_completions")
        if complist:
            s += complist[0]

        log.debug("printing: '%s'", s[-1 * (maxx - self.minx):])
        try:
            self.pad.addstr(s[-1 * (maxx - self.minx):])
        except:
            pass
        self.x = self.pad.getyx()[1]
        self.pad.clrtoeol()
        self.pad.move(0, min(maxx - 1, self.minx + get_rlpoint()))
        try:
            self.callbacks["refresh"]()
        except:
            pass

    def redraw(self):
        pass

    def is_input(self):
        return True

    def get_opt_name(self):
        return "input"

    def get_height(self, mheight):
        return 1

    def get_width(self, mwidth):
        return mwidth



# Clean Title Plugin
# by Jack Miller
# v1.0

# This plugin will strip some annoying content out of story titles. Even though
# most of it should technically still be in there (we already try to parse
# HTML, etc.) some feeds are poorly defined and will double escape HTML and
# insert annoying newlines.

# For now, just do a simple string replace, regexen are probably a bit heavy
# for this.

replacements = [
    ("\n", ""),
    ("<em>",""),
    ("</em>",""),
    ("<strong>",""),
    ("</strong>",""),
    ("<nobr />", ""),
]

# Also included, the option to forcibly remove anything that looks like HTML.
# Useful for cleaning horribly formatted feeds, but possibly destructive of
# good content wrapped in <>. Set this to True to enable.

NO_HTML_EVER = False

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_curses.story import StoryPlugin

# From kjellgren (canto-curses issue #18)

def remove_html_markup(s):
    tag = False
    quote = False
    out = ""
    for c in s:
        if c == '<' and not quote:
            tag = True
        elif c == '>' and not quote:
            tag = False
        elif (c == '"' or c == "'") and tag:
            quote = not quote
        elif not tag:
            out = out + c
    return out

class CleanTitle(StoryPlugin):
    def __init__(self, story):
        self.story = story
        self.plugin_attrs = { "edit_clean" : self.edit_clean }

    def edit_clean(self):
        t = self.story.content["title"]

        for o,n in replacements:
            t = t.replace(o, n)

        if NO_HTML_EVER:
            t = remove_html_markup(t)

        self.story.content["title"] = t
